#![feature(int_roundings)]
#![deny(unsafe_op_in_unsafe_fn)]

use std::ffi::CString;
use std::os::mikros::loader::Loader;
use std::os::mikros::syscalls::{get_initrd, new_process};
use std::path::Path;
use std::process::Command;
use tar_no_std::TarArchiveRef;

fn main() {
    syslog_msg_ipc::register_callback();
    let initrd = TarArchiveRef::new(get_initrd()).unwrap();
    let syslog_pid = run_initrd_proc(&initrd, "bin/syslog");
    let syslog_client = syslog_rpc::Client::new(syslog_pid);
    syslog_client.subscribe_to_binary("devfs".to_string(), vec![0, 1]);
    syslog_client.subscribe_to_binary("tarfs".to_string(), vec![0]);
    syslog_client.send_text_message("init", "Started syslog").unwrap();
    let vfs_pid = run_initrd_proc(&initrd, "bin/vfs");
    syslog_client.send_text_message("init", "Started the VFS").unwrap();
    run_initrd_proc(&initrd, "bin/devfs");
    syslog_client.send_text_message("init", "Started the devfs").unwrap();
    syslog_msg_ipc::get_syslog_msg();
    let vfs_client = vfs_rpc::Client::new(vfs_pid);
    vfs_client.mount(Path::new("/dummy"), "devfs", Path::new("/dev")).unwrap();
    syslog_client.send_text_message("init", "Mounted /dev").unwrap();
    run_initrd_proc(&initrd, "bin/tarfs");
    syslog_client.send_text_message("init", "Started the tar archive FS").unwrap();
    loop {
        let msg = syslog_msg_ipc::get_syslog_msg();
        if msg.from == "tarfs" {
            break;
        }
    }
    run_initrd_proc(&initrd, "bin/initrd_driver");
    syslog_client.send_text_message("init", "Started the initrd pseudodevice driver").unwrap();
    loop {
        let msg = syslog_msg_ipc::get_syslog_msg();
        if msg.from == "devfs" {
            if let Some(binary) = msg.binary {
                if binary.kind == 1 && std::str::from_utf8(binary.data.as_slice()).unwrap() == "initrd" {
                    break;
                }
            }
        }
    }
    vfs_client.mount(Path::new("/dev/initrd"), "tarfs", Path::new("/")).unwrap();
    syslog_client.send_text_message("init", "Mounted the initrd as the root FS").unwrap();
    Command::new("/bin/load_test").spawn().unwrap();
}

fn run_initrd_proc(initrd: &TarArchiveRef, path: &str) -> u64 {
    let proc =
        initrd
            .entries()
            .find(|entry| entry.filename().as_str().unwrap() == path)
            .expect(&format!("{} not found", path))
            .data();
    let (space, entry) = Loader::load(&proc);
    let path_arg = CString::new(path).unwrap();
    new_process(entry as _, space, &[path_arg.as_bytes_with_nul()]).unwrap()
}
